#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <math.h>
#include <stdbool.h>
#include <ncurses.h>
#include <unistd.h>
#include <termios.h>
#include <limits.h>
#include <time.h>
#include <SDL2/SDL.h>
#include <SDL2/SDL_image.h>

#include "consts.h"
#include "base.h"
#include "display.h"
#include "edit.h"
#include "assembly.h"

cell* start ;
cell* current ;

int* item_costs ;
int* item_start ;
int* item_end ;
bool* item_is_normal_type ;
bool* item_is_special_type ;

int* cell_count ;
cell** playerCells ;

player* players ;

imgs digits ;
imgs letters ;

void init_empty_board() {
    start = malloc(sizeof(cell));
    start->id = 0 ;
    start->unique_id = 1 ;
    start->coord_x = 0 ;
    start->coord_y = 0 ;
    start->star_allowed = false ;
    start->type = START ;
    start->prev_1 = NULL;
    start->prev_2 = NULL;
    start->prev_3 = NULL;
    start->prev_4 = NULL;
    start->next_1 = NULL;
    start->next_2 = NULL;
    start->next_3 = NULL;
    start->next_4 = NULL;
    start->meta = "start";

    current = start ;

    cell_count = malloc(sizeof(int)*n_cell_type);
    cell_count[0] = 1;
    for(int i = 1; i < n_cell_type; i++) {
        cell_count[i] = 0;
    }
}

void init_players() {
    players = malloc(sizeof(player)*n_players);
    for(int i = 0; i < n_players; i++) {
        players[i].name = "player" ;
        players[i].standing_tile = start ;
        players[i].coins = 10 ;
        players[i].stars = 0 ;
        players[i].items = malloc(sizeof(ITEM)*3) ;
        players[i].max_items = 3 ;
        players[i].item_len = 0 ;

        players[i].stats.minigames_wins = 0;
        players[i].stats.coin_collected = 0 ;
        players[i].stats.red_count = 0 ;
        players[i].stats.bowser_count = 0 ;
        players[i].stats.items_used = 0 ;
        players[i].stats.lost_coins = 0 ;
        players[i].stats.lost_stars = 0 ;
        players[i].stats.tiles_travelled = 0 ;
        players[i].stats.events_triggered = 0 ;
        players[i].stats.hidden_blocks = 0 ;
    }
}

void init_shop(char* filename) {
    item_costs = malloc(sizeof(int)*n_items);
    item_start = malloc(sizeof(int)*n_items) ;
    item_end = malloc(sizeof(int)*n_items) ;
    item_is_normal_type = malloc(sizeof(bool)*n_items) ;
    item_is_special_type = malloc(sizeof(bool)*n_items) ;

    FILE* ptr = fopen(filename, "r");

    if(ptr == NULL) {
        fprintf(stderr, "ERROR : unable to load config file %s\n", filename);
        exit(1);
    } else {
        printf("    Loading %s...\n", filename);
    }

    int buffer = -1 ;
    int index = 0 ;
    int sub_index = 0 ;

    char c = fgetc(ptr);

    bool is_reading = false ;
    bool ignore_line = false ;

    while(c != EOF && c != '$') {
        if(!ignore_line) {
            if(c == '{') {
                is_reading = true;
            } else if(c == '}') {
                is_reading = false;
                sub_index = 0;
            } else if(c == '*') {
                ignore_line = true;
            } else if(c == ':') {
                buffer = 0;
            } else if(is_reading && buffer != -1) {
                if(is_an_integer(c)) {
                    buffer *= 10;
                    buffer += (char)c - 48; 
                } else {
                    if(sub_index == 0) {
                        index = buffer;
                    } else if(sub_index == 1) {
                        item_start[index] = buffer;
                    } else if(sub_index == 2) {
                        item_end[index] = n_turns - buffer;
                    } else if(sub_index == 3) {
                        item_costs[index] = buffer;
                    } else if(sub_index == 4) {
                        item_is_normal_type[index] = (bool)buffer;
                    } else if(sub_index == 5) {
                        item_is_special_type[index] = (bool)buffer;
                    };
                    sub_index += 1;
                    buffer = -1;
                }
            }
        };
        if(c == '\n') {
            ignore_line = false;
        };
        c = fgetc(ptr);
    };
    fclose(ptr);
}

void import_digits(SDL_Renderer* renderer) {
    imgs res;
    res.arr = malloc(sizeof(SDL_Texture*)*10);
    SDL_Texture* texture;
    SDL_Surface* img;

    // -------------------------------------------------------- //
    img = SDL_LoadBMP("./res/digits/digit-0.bmp");
    texture = SDL_CreateTextureFromSurface(renderer, img);
    SDL_FreeSurface(img);
    res.arr[0] = texture;

    // -------------------------------------------------------- //
    img = SDL_LoadBMP("./res/digits/digit-1.bmp");
    texture = SDL_CreateTextureFromSurface(renderer, img);
    SDL_FreeSurface(img);
    res.arr[1] = texture;

    // -------------------------------------------------------- //
    img = SDL_LoadBMP("./res/digits/digit-2.bmp");
    texture = SDL_CreateTextureFromSurface(renderer, img);
    SDL_FreeSurface(img);
    res.arr[2] = texture;

    // -------------------------------------------------------- //
    img = SDL_LoadBMP("./res/digits/digit-3.bmp");
    texture = SDL_CreateTextureFromSurface(renderer, img);
    SDL_FreeSurface(img);
    res.arr[3] = texture;

    // -------------------------------------------------------- //
    img = SDL_LoadBMP("./res/digits/digit-4.bmp");
    texture = SDL_CreateTextureFromSurface(renderer, img);
    SDL_FreeSurface(img);
    res.arr[4] = texture;

    // -------------------------------------------------------- //
    img = SDL_LoadBMP("./res/digits/digit-5.bmp");
    texture = SDL_CreateTextureFromSurface(renderer, img);
    SDL_FreeSurface(img);
    res.arr[5] = texture;

    // -------------------------------------------------------- //
    img = SDL_LoadBMP("./res/digits/digit-6.bmp");
    texture = SDL_CreateTextureFromSurface(renderer, img);
    SDL_FreeSurface(img);
    res.arr[6] = texture;

    // -------------------------------------------------------- //
    img = SDL_LoadBMP("./res/digits/digit-7.bmp");
    texture = SDL_CreateTextureFromSurface(renderer, img);
    SDL_FreeSurface(img);
    res.arr[7] = texture;

    // -------------------------------------------------------- //
    img = SDL_LoadBMP("./res/digits/digit-8.bmp");
    texture = SDL_CreateTextureFromSurface(renderer, img);
    SDL_FreeSurface(img);
    res.arr[8] = texture;

    // -------------------------------------------------------- //
    img = SDL_LoadBMP("./res/digits/digit-9.bmp");
    texture = SDL_CreateTextureFromSurface(renderer, img);
    SDL_FreeSurface(img);
    res.arr[9] = texture;

    res.len = 10 ;
    digits = res;
}

void import_letters(SDL_Renderer* renderer) {
    imgs res;
    res.arr = malloc(sizeof(SDL_Texture*)*26);
    SDL_Texture* texture;
    SDL_Surface* img;

    int cc = 0 ;

    // -------------------------------------------------------- //
    img = SDL_LoadBMP("./res/letters/letter-a.bmp");
    texture = SDL_CreateTextureFromSurface(renderer, img);
    SDL_FreeSurface(img);
    res.arr[cc] = texture;
    cc += 1 ;
    // -------------------------------------------------------- //
    img = SDL_LoadBMP("./res/letters/letter-b.bmp");
    texture = SDL_CreateTextureFromSurface(renderer, img);
    SDL_FreeSurface(img);
    res.arr[cc] = texture;
    cc += 1 ;
    // -------------------------------------------------------- //
    img = SDL_LoadBMP("./res/letters/letter-c.bmp");
    texture = SDL_CreateTextureFromSurface(renderer, img);
    SDL_FreeSurface(img);
    res.arr[cc] = texture;
    cc += 1 ;
    // -------------------------------------------------------- //
    img = SDL_LoadBMP("./res/letters/letter-d.bmp");
    texture = SDL_CreateTextureFromSurface(renderer, img);
    SDL_FreeSurface(img);
    res.arr[cc] = texture;
    cc += 1 ;
    // -------------------------------------------------------- //
    img = SDL_LoadBMP("./res/letters/letter-e.bmp");
    texture = SDL_CreateTextureFromSurface(renderer, img);
    SDL_FreeSurface(img);
    res.arr[cc] = texture;
    cc += 1 ;
    // -------------------------------------------------------- //
    img = SDL_LoadBMP("./res/letters/letter-f.bmp");
    texture = SDL_CreateTextureFromSurface(renderer, img);
    SDL_FreeSurface(img);
    res.arr[cc] = texture;
    cc += 1 ;
    // -------------------------------------------------------- //
    img = SDL_LoadBMP("./res/letters/letter-g.bmp");
    texture = SDL_CreateTextureFromSurface(renderer, img);
    SDL_FreeSurface(img);
    res.arr[cc] = texture;
    cc += 1 ;
    // -------------------------------------------------------- //
    img = SDL_LoadBMP("./res/letters/letter-h.bmp");
    texture = SDL_CreateTextureFromSurface(renderer, img);
    SDL_FreeSurface(img);
    res.arr[cc] = texture;
    cc += 1 ;
    // -------------------------------------------------------- //
    img = SDL_LoadBMP("./res/letters/letter-i.bmp");
    texture = SDL_CreateTextureFromSurface(renderer, img);
    SDL_FreeSurface(img);
    res.arr[cc] = texture;
    cc += 1 ;
    // -------------------------------------------------------- //
    img = SDL_LoadBMP("./res/letters/letter-j.bmp");
    texture = SDL_CreateTextureFromSurface(renderer, img);
    SDL_FreeSurface(img);
    res.arr[cc] = texture;
    cc += 1 ;
    // -------------------------------------------------------- //
    img = SDL_LoadBMP("./res/letters/letter-k.bmp");
    texture = SDL_CreateTextureFromSurface(renderer, img);
    SDL_FreeSurface(img);
    res.arr[cc] = texture;
    cc += 1 ;
    // -------------------------------------------------------- //
    img = SDL_LoadBMP("./res/letters/letter-l.bmp");
    texture = SDL_CreateTextureFromSurface(renderer, img);
    SDL_FreeSurface(img);
    res.arr[cc] = texture;
    cc += 1 ;
    // -------------------------------------------------------- //
    img = SDL_LoadBMP("./res/letters/letter-m.bmp");
    texture = SDL_CreateTextureFromSurface(renderer, img);
    SDL_FreeSurface(img);
    res.arr[cc] = texture;
    cc += 1 ;
    // -------------------------------------------------------- //
    img = SDL_LoadBMP("./res/letters/letter-n.bmp");
    texture = SDL_CreateTextureFromSurface(renderer, img);
    SDL_FreeSurface(img);
    res.arr[cc] = texture;
    cc += 1 ;
    // -------------------------------------------------------- //
    img = SDL_LoadBMP("./res/letters/letter-o.bmp");
    texture = SDL_CreateTextureFromSurface(renderer, img);
    SDL_FreeSurface(img);
    res.arr[cc] = texture;
    cc += 1 ;
    // -------------------------------------------------------- //
    img = SDL_LoadBMP("./res/letters/letter-p.bmp");
    texture = SDL_CreateTextureFromSurface(renderer, img);
    SDL_FreeSurface(img);
    res.arr[cc] = texture;
    cc += 1 ;
    // -------------------------------------------------------- //
    img = SDL_LoadBMP("./res/letters/letter-q.bmp");
    texture = SDL_CreateTextureFromSurface(renderer, img);
    SDL_FreeSurface(img);
    res.arr[cc] = texture;
    cc += 1 ;
    // -------------------------------------------------------- //
    img = SDL_LoadBMP("./res/letters/letter-r.bmp");
    texture = SDL_CreateTextureFromSurface(renderer, img);
    SDL_FreeSurface(img);
    res.arr[cc] = texture;
    cc += 1 ;
    // -------------------------------------------------------- //
    img = SDL_LoadBMP("./res/letters/letter-s.bmp");
    texture = SDL_CreateTextureFromSurface(renderer, img);
    SDL_FreeSurface(img);
    res.arr[cc] = texture;
    cc += 1 ;
    // -------------------------------------------------------- //
    img = SDL_LoadBMP("./res/letters/letter-t.bmp");
    texture = SDL_CreateTextureFromSurface(renderer, img);
    SDL_FreeSurface(img);
    res.arr[cc] = texture;
    cc += 1 ;
    // -------------------------------------------------------- //
    img = SDL_LoadBMP("./res/letters/letter-u.bmp");
    texture = SDL_CreateTextureFromSurface(renderer, img);
    SDL_FreeSurface(img);
    res.arr[cc] = texture;
    cc += 1 ;
    // -------------------------------------------------------- //
    img = SDL_LoadBMP("./res/letters/letter-v.bmp");
    texture = SDL_CreateTextureFromSurface(renderer, img);
    SDL_FreeSurface(img);
    res.arr[cc] = texture;
    cc += 1 ;
    // -------------------------------------------------------- //
    img = SDL_LoadBMP("./res/letters/letter-w.bmp");
    texture = SDL_CreateTextureFromSurface(renderer, img);
    SDL_FreeSurface(img);
    res.arr[cc] = texture;
    cc += 1 ;
    // -------------------------------------------------------- //
    img = SDL_LoadBMP("./res/letters/letter-x.bmp");
    texture = SDL_CreateTextureFromSurface(renderer, img);
    SDL_FreeSurface(img);
    res.arr[cc] = texture;
    cc += 1 ;
    // -------------------------------------------------------- //
    img = SDL_LoadBMP("./res/letters/letter-y.bmp");
    texture = SDL_CreateTextureFromSurface(renderer, img);
    SDL_FreeSurface(img);
    res.arr[cc] = texture;
    cc += 1 ;
    // -------------------------------------------------------- //
    img = SDL_LoadBMP("./res/letters/letter-z.bmp");
    texture = SDL_CreateTextureFromSurface(renderer, img);
    SDL_FreeSurface(img);
    res.arr[cc] = texture;

    res.len = 26 ;
    letters = res;
}

void free_digits(imgs dgts) {
    for(int i = 0; i < dgts.len; i++) {
        SDL_DestroyTexture(dgts.arr[i]);
    }
    free(dgts.arr);
}

void init_everything(SDL_Renderer* renderer) {
    printf("Initializing board...\n");
    init_empty_board();
    printf("Initializing players...\n");
    init_players();
    printf("Initializing items...\n");
    init_shop("src/config.txt");
    printf("Initializing digits...\n");
    import_digits(renderer);
    import_letters(renderer);
    printf("Initialization complete!\n\n");
    //print_items();
}

void print_items() {
    for(int i = 0; i < n_items; i++) {
        printf("Item : %d\n", i);
        printf("cost : %d\nTurns : [%d - %d]\ntypes=[N = %d , S = %d]", item_costs[i], item_start[i], item_end[i], item_is_normal_type[i], item_is_special_type[i]);
        printf("\n\n");
    }
}

void destroy_everything() {
    printf("Destroying everything...\n");
    for(int i = 0; i < n_players; i++) {
        free(players[i].items);
    };
    free(players);
    free(item_costs);
    free(item_start);
    free(item_end);
    free(item_is_normal_type);
    free(item_is_special_type);
    free(cell_count);

    free_digits(digits);
    free_digits(letters);

    if(n_loaded != -1) {
        free(all_loaded_cells);
    };

    printf("Destruction complete!...\n\n");
}

#include "consts.h"
#include "assembly.h"